@@ -16,21 +16,25 @@ import json
 
 import requests
 
+import time
+
 from slacker.utils import get_item_id_by_name
 
 
-__version__ = '0.9.60'
+__version__ = '0.12.0'
 
 API_BASE_URL = 'https://slack.com/api/{api}'
 DEFAULT_TIMEOUT = 10
-
+DEFAULT_RETRIES = 0
+# seconds to wait after a 429 error if Slack's API doesn't provide one
+DEFAULT_WAIT = 20
 
 __all__ = ['Error', 'Response', 'BaseAPI', 'API', 'Auth', 'Users', 'Groups',
            'Channels', 'Chat', 'IM', 'IncomingWebhook', 'Search', 'Files',
            'Stars', 'Emoji', 'Presence', 'RTM', 'Team', 'Reactions', 'Pins',
            'UserGroups', 'UserGroupsUsers', 'MPIM', 'OAuth', 'DND', 'Bots',
            'FilesComments', 'Reminders', 'TeamProfile', 'UsersProfile',
-           'IDPGroups', 'Apps', 'AppsPermissions', 'Slacker']
+           'IDPGroups', 'Apps', 'AppsPermissions', 'Slacker', 'Dialog']
 
 
 class Error(Exception):
@@ -50,22 +54,45 @@ class Response(object):
 
 class BaseAPI(object):
     def __init__(self, token=None, timeout=DEFAULT_TIMEOUT, proxies=None,
-                 session=None):
+                 session=None, rate_limit_retries=DEFAULT_RETRIES):
         self.token = token
         self.timeout = timeout
         self.proxies = proxies
         self.session = session
+        self.rate_limit_retries = rate_limit_retries
 
     def _request(self, method, api, **kwargs):
         if self.token:
             kwargs.setdefault('params', {})['token'] = self.token
 
-        response = method(API_BASE_URL.format(api=api),
-                          timeout=self.timeout,
-                          proxies=self.proxies,
-                          **kwargs)
+        # while we have rate limit retries left, fetch the resource and back
+        # off as Slack's HTTP response suggests
+        for retry_num in range(self.rate_limit_retries):
+            response = method(API_BASE_URL.format(api=api),
+                              timeout=self.timeout,
+                              proxies=self.proxies,
+                              **kwargs)
 
-        response.raise_for_status()
+            if response.status_code == requests.codes.ok:
+                break
+
+            # handle HTTP 429 as documented at
+            # https://api.slack.com/docs/rate-limits
+            elif response.status_code == requests.codes.too_many: # HTTP 429
+                time.sleep(int(response.headers.get('retry-after', DEFAULT_WAIT)))
+                continue
+
+            else:
+                response.raise_for_status()
+
+        else:
+            # with no retries left, make one final attempt to fetch the resource,
+            # but do not handle too_many status differently
+            response = method(API_BASE_URL.format(api=api),
+                              timeout=self.timeout,
+                              proxies=self.proxies,
+                              **kwargs)
+            response.raise_for_status()
 
         response = Response(response.text)
         if not response.successful:
@@ -113,6 +140,15 @@ class Auth(BaseAPI):
         return self.post('auth.revoke', data={'test': int(test)})
 
 
+class Dialog(BaseAPI):
+    def open(self, dialog, trigger_id):
+        return self.post('dialog.open',
+                         data={
+                             'dialog': json.dumps(dialog),
+                             'trigger_id': trigger_id,
+                         })
+
+
 class UsersProfile(BaseAPI):
     def get(self, user=None, include_labels=False):
         return super(UsersProfile, self).get(
@@ -157,8 +193,8 @@ class Users(BaseAPI):
     def admin(self):
         return self._admin
 
-    def info(self, user):
-        return self.get('users.info', params={'user': user})
+    def info(self, user, include_locale=False):
+        return self.get('users.info', params={'user': user, 'include_locale': include_locale})
 
     def list(self, presence=False):
         return self.get('users.list', params={'presence': int(presence)})
@@ -321,7 +357,7 @@ class Chat(BaseAPI):
     def post_message(self, channel, text=None, username=None, as_user=None,
                      parse=None, link_names=None, attachments=None,
                      unfurl_links=None, unfurl_media=None, icon_url=None,
-                     icon_emoji=None, thread_ts=None):
+                     icon_emoji=None, thread_ts=None, reply_broadcast=None):
 
         # Ensure attachments are json encoded
         if attachments:
@@ -341,7 +377,8 @@ class Chat(BaseAPI):
                              'unfurl_media': unfurl_media,
                              'icon_url': icon_url,
                              'icon_emoji': icon_emoji,
-                             'thread_ts': thread_ts
+                             'thread_ts': thread_ts,
+                             'reply_broadcast': reply_broadcast
                          })
 
     def me_message(self, channel, text):
@@ -408,6 +445,13 @@ class Chat(BaseAPI):
                              'user_auth_url': user_auth_url,
                          })
 
+    def get_permalink(self, channel, message_ts):
+        return self.get('chat.getPermalink',
+                        params={
+                            'channel': channel,
+                            'message_ts': message_ts
+                        })
+
 
 class IM(BaseAPI):
     def list(self):
@@ -551,7 +595,7 @@ class Files(BaseAPI):
                         params={'file': file_, 'count': count, 'page': page})
 
     def upload(self, file_=None, content=None, filetype=None, filename=None,
-               title=None, initial_comment=None, channels=None):
+               title=None, initial_comment=None, channels=None, thread_ts=None):
         if isinstance(channels, (tuple, list)):
             channels = ','.join(channels)
 
@@ -561,12 +605,20 @@ class Files(BaseAPI):
             'filename': filename,
             'title': title,
             'initial_comment': initial_comment,
-            'channels': channels
+            'channels': channels,
+            'thread_ts': thread_ts
         }
 
         if file_:
-            with open(file_, 'rb') as f:
-                return self.post('files.upload', data=data, files={'file': f})
+            if isinstance(file_, str):
+                with open(file_, 'rb') as f:
+                    return self.post(
+                        'files.upload', data=data, files={'file': f}
+                    )
+
+            return self.post(
+                'files.upload', data=data, files={'file': file_}
+            )
         else:
             return self.post('files.upload', data=data)
 
@@ -656,9 +708,13 @@ class Team(BaseAPI):
     def info(self):
         return self.get('team.info')
 
-    def access_logs(self, count=None, page=None):
+    def access_logs(self, count=None, page=None, before=None):
         return self.get('team.accessLogs',
-                        params={'count': count, 'page': page})
+                        params={
+                            'count': count,
+                            'page': page,
+                            'before': before
+                        })
 
     def integration_logs(self, service_id=None, app_id=None, user=None,
                          change_type=None, count=None, page=None):
@@ -977,7 +1033,7 @@ class Slacker(object):
 
     def __init__(self, token, incoming_webhook_url=None,
                  timeout=DEFAULT_TIMEOUT, http_proxy=None, https_proxy=None,
-                 session=None):
+                 session=None, rate_limit_retries=DEFAULT_RETRIES):
 
         proxies = self.__create_proxies(http_proxy, https_proxy)
         api_args = {
@@ -985,6 +1041,7 @@ class Slacker(object):
             'timeout': timeout,
             'proxies': proxies,
             'session': session,
+            'rate_limit_retries': rate_limit_retries,
         }
         self.im = IM(**api_args)
         self.api = API(**api_args)
@@ -994,6 +1051,7 @@ class Slacker(object):
         self.auth = Auth(**api_args)
         self.bots = Bots(**api_args)
         self.chat = Chat(**api_args)
+        self.dialog = Dialog(**api_args)
         self.team = Team(**api_args)
         self.pins = Pins(**api_args)
         self.mpim = MPIM(**api_args)
